freeslot("mt_hitbox","spr_htbx","s_hitbox")
mobjinfo[MT_HITBOX] = {
        spawnstate = S_HITBOX,
        flags = MF_SCENERY|MF_NOTHINK|MF_NOBLOCKMAP
}

states[S_HITBOX] = {
        sprite = SPR_HTBX,
        frame = A,
		tics = 1,
        nextstate = S_HITBOX
}

local CV_Density = CV_RegisterVar{
	name = "hitbox_density",
	defaultvalue = 4,
	flags = CV_NETVAR,
	PossibleValue = {MIN = 0, MAX = 10}
}

local CV_Rings = CV_RegisterVar{
	name = "hitbox_rings",
	defaultvalue = 1,
	flags = CV_NETVAR,
	PossibleValue = {MIN = 0, MAX = 1}
}

local CV_Player = CV_RegisterVar{
	name = "hitbox_player",
	defaultvalue = 1,
	flags = CV_NETVAR,
	PossibleValue = {MIN = 0, MAX = 1}
}

local CV_Color = CV_RegisterVar{
	name = "hitbox_color",
	defaultvalue = 1,
	flags = CV_NETVAR,
	PossibleValue = {MIN = 0, MAX = 1}
}

local CV_Distance = CV_RegisterVar{
	name = "hitbox_distance",
	defaultvalue = 512,
	flags = CV_NETVAR,
	PossibleValue = {MIN = 0, MAX = 1280}
}

local Hitboxes = {}

local Vertices = {}

local DensityCheck = CV_Density.value

local CanSeeHitbox = function(mo)
	local dist = R_PointToDist(mo.x,mo.y)
	if dist > CV_Distance.value*FRACUNIT then return false end
	if mo.type == MT_PLAYER and dist < FRACUNIT*72 then return false end
	return true
end

local DoVertex = function(mo)
	//Object does not exist anymore
	if not(mo and mo.valid) then return nil end
	//Target does not exist anymore
	local target = mo.target
	if not(target and target.valid) then
		P_RemoveMobj(mo)
	return nil end
	//Debug hitboxes are turned off
	if not(CV_Density.value) then
		P_RemoveMobj(mo)
	return nil end
	//Ring hitboxes are turned off
	if target.type == MT_RING and not(CV_Rings.value) then
		P_RemoveMobj(mo)
	return nil end
	//Player hitboxes are turned off
	if target.type == MT_PLAYER and not(CV_Player.value) then
		P_RemoveMobj(mo)
	return nil end
	//Density value has changed since the last frame
	if CV_Density.value != DensityCheck then
		P_RemoveMobj(mo)
	return nil end	
	//Hitbox is too far away
	if not(CanSeeHitbox(target)) then
		P_RemoveMobj(mo)
	return nil end
	//All checks passed. Updating position
	P_TeleportMove(mo,target.x+FixedMul(mo.xo,target.scale),target.y+FixedMul(mo.yo,target.scale),target.z+FixedMul(mo.zo,target.scale))
	target.vertices_exist = true //Confirming to our target object that vertices still exist

	//Do color coding
	if not(CV_Color.value) then
		mo.color = SKINCOLOR_JET
	elseif target.player then
		mo.color = SKINCOLOR_BLUE
	elseif target.flags&(MF_PAIN|MF_MISSILE) then
		mo.color = SKINCOLOR_RED
	elseif target.flags&MF_SOLID then
		mo.color = SKINCOLOR_GREY
	elseif target.flags&MF_PUSHABLE then
		mo.color = SKINCOLOR_PURPLE
	elseif target.flags&(MF_ENEMY|MF_BOSS) then
		mo.color = SKINCOLOR_ORANGE
	elseif target.flags&(MF_SPECIAL|MF_SPRING) then
		mo.color = SKINCOLOR_YELLOW
	else
		mo.color = SKINCOLOR_GREEN
	end
	return mo
end

local SpawnVertex = function(mo,xo,yo,zo,scale)
	if not(mo and mo.valid) then return end
	local v = P_SpawnMobj(mo.x,mo.y,mo.z,MT_HITBOX)
	if not(v and v.valid) then return end
	if scale == nil then scale = FRACUNIT*3/4 end
	v.scale = scale
	v.target = mo
	v.xo = xo
	v.yo = yo
	v.zo = zo
	Vertices[#Vertices+1] = v
	DoVertex(v)
end

local FixedLerp = function(val1,val2,amt)
	local p = FixedMul(FRACUNIT-amt,val1) + FixedMul(amt,val2)
	return p
end

local DrawLine = function(mo,x1,x2,y1,y2,z1,z2,axis)
	local mult = FRACUNIT*48
	local density = CV_Density.value
	if axis == 0 then
		density = mo.radius*$/mult
	else
		density = mo.height*$/mult
	end
	for n = 0, density
		if n == 0 or n == density then continue end
		local x = FixedLerp(x1,x2,FRACUNIT*n/density)
		local y = FixedLerp(y1,y2,FRACUNIT*n/density)
		local z = FixedLerp(z1,z2,FRACUNIT*n/density)
		local s
		if n == density/2 and not(density&1) then
			s = FRACUNIT/2
		else
			s = FRACUNIT/4
		end
		SpawnVertex(mo,x,y,z,s)
	end
end

local UpdateHitbox = function(mo)
	if not(mo and mo.valid) then return nil end
	if mo.type == MT_RING and not(CV_Rings.value) then return true end
	if mo.vertices_exist then
		mo.vertices_exist = false //Reset the check for next frame
	return true end
	if not(CanSeeHitbox(mo)) then return true end
	local r = mo.radius
	local h = mo.height
	local d = CV_Density.value
	SpawnVertex(mo,r,r,0)	
	SpawnVertex(mo,-r,r,0)
	SpawnVertex(mo,r,-r,0)
	SpawnVertex(mo,-r,-r,0)
	SpawnVertex(mo,r,r,h)
	SpawnVertex(mo,-r,r,h)
	SpawnVertex(mo,r,-r,h)
	SpawnVertex(mo,-r,-r,h)
	if mo.type == MT_RING then return true end
	
	if d > 1 then
		DrawLine(mo,-r,r,r,r,0,0,0)
		DrawLine(mo,-r,r,-r,-r,0,0,0)
		DrawLine(mo,-r,r,r,r,h,h,0)
		DrawLine(mo,-r,r,-r,-r,h,h,0)
		DrawLine(mo,r,r,-r,r,0,0,0)
		DrawLine(mo,-r,-r,-r,r,0,0,0)
		DrawLine(mo,r,r,-r,r,h,h,0)
		DrawLine(mo,-r,-r,-r,r,h,h,0)
		DrawLine(mo,r,r,r,r,0,h,1)
		DrawLine(mo,-r,-r,r,r,0,h,1)
		DrawLine(mo,r,r,-r,-r,0,h,1)
		DrawLine(mo,-r,-r,-r,-r,0,h,1)
	end
	
	return true
end

addHook("MobjSpawn",function(mo)
	if not(mo.type == MT_PLAYER or mo.flags&(MF_SPECIAL|MF_MISSILE|MF_PAIN|MF_ENEMY|MF_BOSS|MF_SOLID|MF_PUSHABLE|MF_SPRING))
	then return end
	Hitboxes[#Hitboxes+1] = mo
	mo.vertices_exist = false
end,MT_NULL)

addHook("PostThinkFrame",function()
	//Update vertex positions
	local queue = {}
	for n = 1,#Vertices
		queue[#queue+1] = Vertices[n]
	end
	Vertices = {}
	local size = #queue
	for n = 1, size
		local vertex = queue[n]
		if DoVertex(vertex) then
			Vertices[#Vertices+1] = vertex
		end
	end
	
	//Update DensityCheck value
	DensityCheck = CV_Density.value
	
	//Update hitboxes
	queue = {}
	for n = 1,#Hitboxes
		queue[#queue+1] = Hitboxes[n]
	end
	Hitboxes = {}
	local size = #queue
	for n = 1, size
		local hitbox = queue[n]
		if UpdateHitbox(hitbox) then
			Hitboxes[#Hitboxes+1] = hitbox
		end
	end
end)

addHook("MapChange",function()
	Vertices = {}
	Hitboxes = {}
end)

